theory ListAndTrees 
  imports Main "HOL-Library.Multiset"
begin

datatype 'a List =
  Empty | 
  Cons 'a "'a List" (infixr "##" 100)

term "(2::nat) ## 3 ## Empty"

subsection \<open>append\<close>

primrec append :: "'a List \<Rightarrow> 'a List \<Rightarrow> 'a List" (infixl "@@" 100) where
  "Empty @@ ys = ys"
| "(x ## xs) @@ ys = x ## (xs @@ ys)"

value "((2::nat) ## 3 ## Empty) @@ (1 ## Empty)"

lemma append_empty_right [simp]:
  "xs @@ Empty = xs"
  by (induction xs) auto

lemma append_assoc [simp]:
  "(xs @@ ys) @@ zs = xs @@ (ys @@ zs)"
  by (induction xs) auto

subsection \<open>reverse\<close>

primrec reverse :: "'a List \<Rightarrow> 'a List" where
  "reverse Empty = Empty"
| "reverse (x ## xs) = reverse xs @@ (x ## Empty)" 

lemma reverse_append [simp]:
 "reverse (xs @@ ys) = reverse ys @@ reverse xs"
  by (induction xs) auto

lemma "reverse (reverse xs) = xs"
  by (induction xs) auto

subsection \<open>reverse_opt\<close>

primrec reverse_opt' where
  "reverse_opt' Empty acc = acc"
| "reverse_opt' (x ## xs) acc = reverse_opt' xs (x ## acc)"

definition reverse_opt where
  "reverse_opt xs = reverse_opt' xs Empty"

lemma reverse_opt':
  "reverse_opt' xs acc = reverse xs @@ acc"
  by (induction xs arbitrary: acc) auto

lemma reverse_opt:
  "reverse_opt xs = reverse xs"
  using reverse_opt'[of xs]
  unfolding reverse_opt_def
  by simp

subsection \<open>insert\<close>

primrec insert :: "nat \<Rightarrow> nat list \<Rightarrow> nat list" where
  "insert a [] = [a]" 
| "insert a (x # xs) = 
     (if a \<le> x then 
          a # x # xs 
      else
          x # insert a xs)" 

lemma set_insert [simp]:
  shows "set (insert a xs) = set xs \<union> {a}"
  by (induction xs) auto

lemma multiset_insert [simp]:
  shows "mset (insert a xs) = {#a#} + mset xs"
  by (induction xs) auto

lemma sorted_insert [simp]:
  assumes "sorted xs"
  shows "sorted (insert a xs)"
  using assms
  by (induction xs) auto

primrec insertion_sort :: "nat list \<Rightarrow> nat list" where
  "insertion_sort [] = []"
| "insertion_sort (x # xs) = insert x (insertion_sort xs)"

lemma sorted_insertion_sort:
  shows "sorted (insertion_sort xs)"
  by (induction xs) auto

lemma mset_insertion_sort:
  shows "mset (insertion_sort xs) = mset xs"
  by (induction xs) auto

section \<open>Trees\<close>

datatype Tree = 
  Nil |
  Node nat Tree Tree

primrec sum_tree :: "Tree \<Rightarrow> nat" where
  "sum_tree Nil = 0"
| "sum_tree (Node x l r) = sum_tree l + x + sum_tree r"

primrec tree_infix :: "Tree \<Rightarrow> nat list" where
  "tree_infix Nil = []"
| "tree_infix (Node x l r) = tree_infix l @ [x] @ tree_infix r"

primrec tree_set :: "Tree \<Rightarrow> nat set" where
  "tree_set Nil = {}"
| "tree_set (Node x l r) = tree_set l \<union> {x} \<union> tree_set r"

primrec tree_mset :: "Tree \<Rightarrow> nat multiset" where
  "tree_mset Nil = {#}"
| "tree_mset (Node x l r) = tree_mset l + {# x #} + tree_mset r"
                            
primrec is_bst :: "Tree \<Rightarrow> bool" where
  "is_bst Nil = True"
| "is_bst (Node x l r) \<longleftrightarrow> is_bst l \<and> is_bst r \<and>
   (\<forall> x' \<in> tree_set l. x' \<le> x) \<and> (\<forall> x' \<in> tree_set r. x \<le> x')"

lemma set_tree_infix [simp]:
  "set (tree_infix t) = tree_set t"
  by (induction t) auto

lemma mset_tree_infix [simp]:
  "mset (tree_infix t) = tree_mset t"
  by (induction t) auto

lemma sorted_tree_infix [simp]: 
  assumes "is_bst t"
  shows "sorted (tree_infix t)"
  using assms
  by (induction t) (force simp add: sorted_append)+

primrec tree_insert :: "nat \<Rightarrow> Tree \<Rightarrow> Tree" where
  "tree_insert a Nil = Node a Nil Nil"
| "tree_insert a (Node x l r) = 
    (if a \<le> x then Node x (tree_insert a l) r                                                
     else Node x l (tree_insert a r))"

lemma set_tree_insert [simp]:
  shows "tree_set (tree_insert a t) = {a} \<union> tree_set t"
  by (induction t) auto

lemma mset_tree_insert [simp]:
  shows "tree_mset (tree_insert a t) = {#a#} + tree_mset t"
  by (induction t) auto

lemma is_bst_tree_insert [simp]:
  assumes "is_bst t"
  shows "is_bst (tree_insert a t)"
  using assms
  by (induction t) auto

definition tree_of_list :: "nat list \<Rightarrow> Tree" where
  "tree_of_list xs = fold tree_insert xs Nil"

lemma is_bst_tree_of_list [simp]:
  shows "is_bst (tree_of_list xs)"
  unfolding tree_of_list_def
  by (induction xs rule: rev_induct) auto

lemma mset_tree_of_list [simp]:
  shows "tree_mset (tree_of_list xs) = mset xs"
  unfolding tree_of_list_def
  by (induction xs rule: rev_induct) auto


definition tree_sort :: "nat list \<Rightarrow> nat list" where
  "tree_sort xs = tree_infix (tree_of_list xs)"

lemma sorted_tree_sort [simp]:
  shows "sorted (tree_sort xs)"
  unfolding tree_sort_def
  by simp

lemma mset_tree_sort [simp]:
  shows "mset (tree_sort xs) = mset xs"
  unfolding tree_sort_def
  by simp

end