theory Locales
  imports Main HOL.Real "HOL-Library.Multiset" GeneralRecursion
begin

locale Geometry = 
  fixes cong :: "'a \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> bool"
  assumes cong_refl: "\<And> x y. cong x y x y"
  assumes cong_id: "\<And> x y z. cong x y z z \<Longrightarrow> x = y"
  assumes cong_trans: "\<And> x y z u v w. \<lbrakk>cong x y z u; cong x y v w\<rbrakk> \<Longrightarrow> cong z u w v" 
begin
lemma cong_comm1:
  assumes "cong x y v w"
  shows "cong y x v w"
  using assms
  using cong_refl cong_trans by blast

lemma cong_comm2:
  assumes "cong x y v w"
  shows "cong x y w v"
  using assms
  using cong_refl cong_trans by blast
end

type_synonym point_R2 = "real \<times> real"

fun dist_sq_R2 :: "point_R2 \<Rightarrow> point_R2 \<Rightarrow> real" where
  "dist_sq_R2 (x1, y1) (x2, y2) = (x2 - x1)^2 + (y2 - y1)^2"

definition cong_R2 :: "point_R2 \<Rightarrow> point_R2 \<Rightarrow> point_R2 \<Rightarrow> point_R2 \<Rightarrow> bool" where
  "cong_R2 x y z w \<longleftrightarrow> dist_sq_R2 x y = dist_sq_R2 z w"

interpretation Geometry_R2: Geometry cong_R2
proof
  fix x y
  show "cong_R2 x y x y"
    by (simp add: cong_R2_def)
next
  fix x y z
  assume "cong_R2 x y z z"
  then show "x = y"
    unfolding cong_R2_def
    by (cases x, cases y, cases z, simp)
next
  fix x y z u v w
  assume "cong_R2 x y z u" "cong_R2 x y v w" 
  then show "cong_R2 z u w v"
    unfolding cong_R2_def
    by (cases x, cases y, cases z, cases u, cases v, cases w, simp add: field_simps power2_eq_square)
qed

section\<open>Merge sort\<close>

locale Split = 
  fixes split :: "nat list \<Rightarrow> (nat list \<times> nat list)"
  assumes split_length: "\<And> xs ys zs. \<lbrakk>length xs > 1; (ys, zs) = split xs\<rbrakk> \<Longrightarrow> length ys < length xs \<and> length zs < length xs"
  assumes split_mset: "\<And> xs ys zs. \<lbrakk>length xs > 1; (ys, zs) = split xs\<rbrakk> \<Longrightarrow> mset xs = mset ys + mset zs"
begin

function (sequential) merge_sort :: "nat list \<Rightarrow> nat list" where
  "merge_sort [] = []"
| "merge_sort [x] = [x]" 
| "merge_sort xs =
    (let (ys, zs) = split xs
      in merge (merge_sort ys) (merge_sort zs))"
  by pat_completeness auto
termination
  apply (relation "measure length")
    apply simp
   apply (metis One_nat_def add_cancel_right_left antisym_conv3 in_measure length_0_conv list.discI list.size(4) not_add_less2 split_length)
  apply (metis One_nat_def in_measure length_0_conv length_Cons list.discI list.size(4) nat.inject not_add_less2 not_less_iff_gr_or_eq split_length)
  done

lemma sorted_merge_sort:
  "sorted (merge_sort xs)"
  by (induction xs rule: merge_sort.induct) (auto split: prod.split)
end

definition split_half :: "nat list \<Rightarrow> (nat list \<times> nat list)" where 
  "split_half xs = 
    (let n = length xs div 2 
      in (take n xs, drop n xs))"

global_interpretation Split_half: Split where
  split = split_half 
  defines merge_sort_half = "Split_half.merge_sort"
proof
  fix xs ys zs
  assume "1 < length xs" "(ys, zs) = split_half xs"
  then show "length ys < length xs \<and> length zs < length xs"
    unfolding split_half_def Let_def
    by auto
next
  fix xs ys zs
  assume "1 < length xs" "(ys, zs) = split_half xs"
  then show "mset xs = mset ys + mset zs"
    unfolding split_half_def Let_def
    by (metis append_take_drop_id old.prod.inject union_code)
qed

value "merge_sort_half [1, 2, 5, 8, 4, 6, 7, 3]"
export_code merge_sort_half in Haskell

end